@@ -8,6 +8,7 @@ from past.builtins import map
 
 from . import blocks, exception, legos, primitives
 from .blocks import Aligned, Block, Checksum, Repeat, Request, REQUESTS, Size
+from .cli import main_helper
 from .connections import (
     BaseSocketConnection,
     FileConnection,
@@ -44,11 +45,13 @@ from .primitives import (
     Bytes,
     Delim,
     DWord,
+    Float,
     FromFile,
     Group,
     Mirror,
     QWord,
     RandomData,
+    Simple,
     Static,
     String,
     Word,
@@ -87,6 +90,7 @@ __all__ = [
     "EventHook",
     "exception",
     "FileConnection",
+    "Float",
     "FromFile",
     "Fuzzable",
     "FuzzableBlock",
@@ -102,6 +106,7 @@ __all__ = [
     "ITargetConnection",
     "legos",
     "LITTLE_ENDIAN",
+    "main_helper",
     "Mirror",
     "MustImplementException",
     "NetworkMonitor",
@@ -136,6 +141,7 @@ __all__ = [
     "s_double",
     "s_dunno",
     "s_dword",
+    "s_float",
     "s_from_file",
     "s_get",
     "s_group",
@@ -156,6 +162,7 @@ __all__ = [
     "s_size",
     "s_sizer",
     "s_static",
+    "s_simple",
     "s_string",
     "s_switch",
     "s_unknown",
@@ -168,6 +175,7 @@ __all__ = [
     "SizerNotUtilizedError",
     "SocketConnection",
     "SSLSocketConnection",
+    "Simple",
     "Static",
     "String",
     "SullyRuntimeError",
@@ -181,7 +189,7 @@ __all__ = [
     "Word",
 ]
 
-__version__ = "0.3.0"
+__version__ = "0.4.0"
 
 
 # REQUEST MANAGEMENT
@@ -693,6 +701,25 @@ def s_static(value=None, name=None):
     blocks.CURRENT.push(Static(name=name, default_value=value))
 
 
+def s_simple(value=None, name=None, fuzz_values=None, fuzzable=True):
+    """
+    Push a "Simple" primitive onto the current block stack. The only mutations will be those specified  in fuzz_values.
+
+    :see: Aliases: s_dunno(), s_raw(), s_unknown()
+
+    :type  value: Raw
+    :param value: Raw static data
+    :type  name:  str
+    :param name:  (Optional, def=None) Specifying a name gives you direct access to a primitive
+    :type  fuzz_values:  list
+    :param fuzz_values:  (Optional, def=None) List of fuzz values.
+    :type  fuzzable: bool
+    :param fuzzable: (Optional, def=True) Enable/disable fuzzing of this primitive
+    """
+
+    blocks.CURRENT.push(Simple(name=name, default_value=value, fuzz_values=fuzz_values, fuzzable=fuzzable))
+
+
 def s_mirror(primitive_name=None, name=None):
     """
     Push a mirror of another primitive onto the current block stack.
@@ -764,6 +791,58 @@ def s_from_file(value="", filename=None, encoding="ascii", fuzzable=True, max_le
     blocks.CURRENT.push(FromFile(name=name, default_value=value, max_len=max_len, filename=filename, fuzzable=fuzzable))
 
 
+def s_float(
+    value=0.0,
+    s_format=".1f",
+    f_min=0.0,
+    f_max=100.0,
+    max_mutations=1000,
+    fuzzable=True,
+    seed=None,
+    encode_as_ieee_754=False,
+    endian="big",
+    name=None,
+):
+    """
+    Push a float onto the current block stack.
+
+    :type  value:              float
+    :param value:              (Optional, def=0.0) Default float value
+    :type  s_format:           str
+    :param s_format:           (Optional, def=".1f") Format of the float value after encoding.
+    :type f_min:               float
+    :param f_min:              (Optional, def=0.0) Min value of generated floats.
+    :type f_max:               float
+    :param f_max:              (Optional, def=100.0) Max value of generated floats.
+    :type max_mutations:       int
+    :param max_mutations:      (Optional, def=1000) Total number of mutations for this individual primitive
+    :type  fuzzable:           bool
+    :param fuzzable:           (Optional, def=True) Enable/disable fuzzing of this primitive
+    :type seed:                int or str or bytes or bytearray
+    :param seed:               (Optional, def=None) Set random.seed() with the given seed for reproducible results
+    :type encode_as_ieee_754:  bool
+    :param encode_as_ieee_754: (Optional, def=False) Enable/disable encoding as IEEE 754 float
+    :type endian:              str
+    :param endian:             (Optional, def='big') Set the endianness for IEEE 754 float
+    :type  name:               str
+    :param name:               (Optional, def=None) Specifying a name gives you direct access to a primitive
+    """
+    blocks.CURRENT.push(
+        Float(
+            name=name,
+            default_value=value,
+            s_format=s_format,
+            f_min=f_min,
+            f_max=f_max,
+            max_mutations=max_mutations,
+            fuzzable=fuzzable,
+            seed=seed,
+            encode_as_ieee_754=encode_as_ieee_754,
+            endian=endian,
+        )
+    )
+
+
 # noinspection PyTypeChecker
 def s_bit_field(
     value=0,
